#!/usr/bin/perl

# [en]	purpose: Perl-based backend for PowerDNS queries
# [de]	Zweck: Perl-Skript zum Beantworten der DNS-Fragen von PowerDNS

# https://doc.powerdns.com/md/appendix/backend-writers-guide/
# https://doc.powerdns.com/md/authoritative/backend-pipe/

# [en]	include this backend in PowerDNS configuration file with these lines:
# [de]	dieses Skript wird mit den folgenden Zeilen in der Konfiguration von
#	PowerDNS aktiviert:
# ---------------------------------------
#  launch=pipe
#  pipe-command=/etc/pdns/pdns_backend.pl
#  query-cache-ttl=0
#  negquery-cache-ttl=0
#  cache-ttl=0
# ---------------------------------------


# [en]	script debugging (the whitespace between fields MUST be a tabulator)
# [de]	Fehlersuche im Skript (die weissen Felder zwischen den Werten muessen
#	Tabulatoren sein)
#
# $ perl /etc/pdns/pdns_backend.pl
# HELO	1
# Q       www.example.net IN      ANY     -1      143.107.0.1 

# [en]	print many messages in PowerDNS logfile
# [de]	Viele Debugmeldung ins Log schreiben?
my $debug = 1;


# [en]	define all CDN nodes in this variable. Mandatory values are identifier
#	(should	be the hostname), an IP address and the country. To make the
#	CDN country-to-node resolution more accurate, specify the values for
#	latitude and longitute of the server
# [de]	Hier werden die einzelnen CDN-Server angegeben. Die Angaben fuer Name
#	(sollte der Hostname sein), IP-Adresse und Land sind zwingend notwendig.
#	Optional (aber sinnvoll	fuer eine praezisere Berechnung) sind die Werte
#	fuer Laengen- und Breitengrad.
my $cdn = {
#	hostname => {
#		ipv4 => '192.0.2.15',
#		ipv6 => '2001:db8::15',
#		country   => 'de',	# ISO 3166 Country Codes
#		# http://dev.maxmind.com/geoip/legacy/codes/iso3166/
#		latitude  => 50.937821,	# [en]	use Google maps to find coordinates
#		longitude => 6.960693,	#	of server location
#					# [de]	Benutze Google Maps um die GPS-Koordinaten
#					#	der Serverstandorte zu ermitteln
#		disabled  => 0,		# [en]	set to 1 for out-of-service
#					# [de]	Fuer Wartungsmodus diesen Wert auf 1 setzen
#	},
	'cdn-de-1' => {
		ipv4 => '192.0.2.49',
		ipv6 => '2001:db8:49::1',
		country   => 'de',
		latitude  => 50.937821,
		longitude => 6.960693,
		disabled  => 0,
	},
	'cdn-cn-1' => {
		ipv4 => '203.0.113.86',
		ipv6 => '2001:db8:86::1',
		country   => 'cn',
		latitude  => 39.902629,
		longitude => 116.396787,
		disabled  => 0,
	},
	'cdn-za-1' => {
		ipv4 => '198.51.100.27',
		ipv6 => '2001:db8:27::1',
		country   => 'za',
		latitude  => -25.745113,
		longitude => 28.220803,
	},
	'cdn-us-1' => {
		ipv4 => '192.0.2.1',
		ipv6 => '2001:db8:1::1',
		country   => 'us',
		latitude  => 33.552402,
		longitude => -86.825203,
		disabled  => 0,
	},
};

# [en]	sometimes the nearest physical distance is not the best connection.
#	Overwrite the coordinate calculation and country selection here
# [de]	Manchmal ist die berechnete Entfernung nicht die Optimale. Hier kann
#	die Zuordnung von Land zu CDN-Server haendisch festgelegt werden.
my $cdn_node_for_country = {
	# [en] country code => cdn_node_name,
	# [de] Landeskuerzel => CDN-Servername,
	br => 'cdn-us-1',
};

# [en]	answer only if client queries for these names
# [de]	Auf welche DNS-Anfrage sollen wir antworten?
my %qnames = (
	'www.example.net'	=> 'ns1.example.net',
	'www-cdn.example.com'	=> 'ns2.example.net',
);

# [en]	if client country is unknown (or something else fails) return this
#	server's IP address
# [de]	Falls das Herkunfsland des Clients nicht gefunden wird, sende seine
#	Anfragen an einen vorgegebenen CDN-Server
my $default_cdn_node = 'cdn-de-1';

# [en]	full-qualified hostname of the DNS server, e.g. ns1.example.net
# 	leave blank to determine the hostname prefixed with 'ns'
# [de]	Kompletter Name des lokalen DNS-Servers, z.B. ns1.example.net
# 	Falls kein Wert angegeben wird, ermittelt das Skript die Domaene und
# 	setzt 'ns' davor.
my $my_fqdn = '';	# 'ns1.example.net';


### [en] no customizing below this line ###
### [de] keine Anpassung unterhalb dieser Zeile ###


use strict;

$|=1;			# [en] no buffering
			# [de] Puffer ausschalten
use constant PI_180 => 3.14159265358979323846264338327950288419716 / 180;

# [en]	receive query (as text) from PowerDNS daemon
# [de]	Empfange die Anfrage des PowerDNS-Dienst
my $line = <>;
chomp( $line );

# [en]	check for correct abi-version (must be version 1)
# [de]	Pruefe die ABI-Version. Dieses Skript erwartet Version 1
if ( $line eq "HELO\t1" ) {
	print "OK GeoCDN backend ready\n";	# [en] print our banner
						# [de] Sende OK und ein Banner
} elsif ( $line =~ /^HELO\s+1/ ) {
	print "OK GeoCDN devel backend ready\n";

} else {
	print "FAIL\n";
	$debug && print STDERR "DEBUG\tReceived '$line'\n";
	<>;
	exit;
}

while ( <> ) {
	$debug && print STDERR "DEBUG\t$$ Received: $_";
	chomp();

	my ( $type,$qname,$qclass,$qtype,$id,$client ) = split( /\s+/, $_, 6 );

	#  https://doc.powerdns.com/md/authoritative/backend-pipe/
	# type       Q
	# qname      www.example.net
	# qclass     IN
	# $qtype     SOA
	# id         -1
	# client-ip  192.168.1.200
    
	if ( $type eq 'Q' ) {	# query    
		# [en]	validate if client is asking for DNS records that we are
		#	willing to answer
		# [de]	Fragt der Client tatsaechlich nach unseren DNS-Eintraegen?
		if ( $qnames{ $qname } ) {
			# e.g.  Q www.example.net IN SOA -1 192.168.1.200
			if ( $qtype eq 'SOA' ) {
				reply( $qname, $qclass, $id, 86400, $qtype,
					$qnames{ $qname } );

			# [en]	powerdns will send all queries as ANY request, so we can't
			# 	know if the client asked for A or AAAA. So let's just provide
			# 	all we have
			# [de]	PowerDNS fragt stets nach ANY Eintraegen, so dass hier nicht
			# 	unterschieden werden kann, ob tatsaechlich A oder AAAA
			# 	gewuenscht ist. Also liefern wir alles was wir haben.
			# e.g.  Q www.example.net IN ANY -1 192.168.1.200
			} elsif ( $qtype eq 'ANY' ) {
				my ( $country, $lat, $long )
					= get_country_for( $client );

				my $cdn_node;
				if ( $country ) {	# resolved IP to country
					$cdn_node = nearest_cdn_node(
						$country, $lat, $long );
				} else {
					# [en] default CDN node is valid record?
					# [de] ist der vorgegebenen CDN-Server
					#      ein gueltiger Eintrag?
					if ( defined $cdn->{$default_cdn_node} ) {
						$cdn_node = $default_cdn_node;
					} else {
						# invalid? pick random record
						# from all available cdn nodes
						my @all_nodes;
						for my $node ( keys %$cdn ) {
							if ( ! $cdn->{ $node }->{'disabled'} ) {
								push @all_nodes, $node;
							}
						}
						my $random_node = rand @all_nodes;
						$cdn_node = $all_nodes[$random_node];

						$debug && print STDERR	"LOG\t".
							"Invalid default node, ".
							"fallback to $cdn_node\n";
					}
				}
		
				# [en]	answer with all known IP addresses
				# [de]	Alle bekannten IP-Adressen zurueckliefern
				reply( $qname, $qclass, abs($id), 3600,
					'A', $cdn->{ $cdn_node }->{'ipv4'} );
				reply( $qname, $qclass, abs($id), 3600,
					'AAAA', $cdn->{ $cdn_node }->{'ipv6'} );
			}
		} else {
			print STDERR "FAIL: client asked for '$qname'\n";
			next;
		}

	} elsif ( $type eq 'AXFR' ) {
		# [en]	we don't support AXFR
		# [de]	keine Antwort bei Fragen nach AXFR
	}
	print "END\n";
}



=item I<reply> I<(query name, query class, ID, TTL, query type, resource record)>

[en] Send the DNS answer to powerdns daemon in the correct format
[de] Formatiere und sende die DNS-Antwort an PowerDNS

=cut
sub reply {
	my ($qname, $qclass, $id, $ttl, $qtype, $rr) = @_;
    
	$debug && print STDERR "DATA\t$qname\t$qclass\t$qtype\t$ttl\t$id\t$rr\n";
	print "DATA\t$qname\t$qclass\t$qtype\t$ttl\t$id\t$rr\n";
	$debug && print STDERR "LOG\treturning $rr for $qname/$qtype\n";
}



=item I<get_country_for> I<(IP address)>

[en] Find the country information (two letter country code) for the give IP address
[de] Suche die Landesinformationen (zwei Zeichen) der angegebenen IP-Adresse

=cut
sub get_country_for( $ ) {
	my ( $client ) = @_;
	$debug && print STDERR "DEBUG\tget_country_for( $client )\n";

	my $cmd = 'geoiplookup';
	if ( $client =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/ ) {	# IPv4
		# [en]	do nothing, command defaults to IP version 4
		# [de]	nichts machen, das Kommando erwartet IPv4
	} else {							# IPv6
		$cmd .= '6';
	}
	$cmd .= " $client";
	$debug && print STDERR "DEBUG\tgeoip-cmd '$cmd'\n";

	for my $line ( `$cmd` ) {
		chomp( $line );

		# [en]	the output of geoiplookup has a variable number
		#	of colums and the latitude/longitude information
		#	is not always in the same colum! E.g.
		# [de]	Die Anzahl der Spalten in der Ausgabe von geoiplookup
		# 	ist unterschiedlich. Also ist die Position von Laengengrad
		# 	und Breitengrad nicht in derselben Spalte! Beispiel: 
		#  DE, 07, Nordrhein-Westfalen, Duesseldorf, 40235, 51.230099, 6.821600, 0, 0
		#  DE, 07, Marienheide, 51709, 51.082802, 7.535100, 0, 0

		if ( $line =~ s/^GeoIP City Edition.*:\s*// ) {
			$debug && print STDERR "DEBUG\t$line\n";
			my @info = split( /\s*,\s*/, $line );
			my ( $country, $lat, $long ) = @info[ 0, -4, -3 ];

			$debug && printf STDERR
				"DEBUG\tcountry:%s lat:%0.2f long:%0.2f\n",
				$country, $lat, $long;
			return(	$country, $lat, $long );
		}
	}
	if ( $debug ) {
		print STDERR "LOG\tError: GeoIP City Edition ".
			"is not available\n";
		print STDERR "LOG\tPlease get the GeoIPCity.dat from MaxMind: ".
			"https://dev.maxmind.com/geoip/legacy/geolite/\n";
	}
	return undef;
}



=item I<nearest_cdn_node> I<(country code, latitude, longitude)>

[en] Calculate the distance between the country and all CDN nodes. Return the nodes
     hostname that is closest to the country.
[de] Berechne die Entfernung zwischen dem Land und jedem CDN-Server. Der Server mit
     der kuerzeste Entfernung gewinnt und wird als Hostname zurueckgeliefert.

=cut
sub nearest_cdn_node( $$$ ) {
	my ( $country, $latitude, $longitude ) = @_;
	$debug && printf STDERR "DEBUG\tnearest_cdn_node( %s, %s, %s )\n",
		$country, $latitude, $longitude;
	my $pref;

	# [en]	static country-to-cdn-node mapping
	# [de]	statische Zuordnung von Land zu CDN-Server
	if ( my $node = $cdn_node_for_country->{ lc $country } ) {
		$debug && printf STDERR ( "DEBUG\ttraffic is mapped to CDN node %s (in %s)\n",
			$node, uc $cdn->{ $node }->{'country'} );

		return $node;
	}

	# [en]	get longitude/latitude about all CDN nodes
	# [de]	ermittle den Laengen-/Breitengrad der CDN-Server
	for my $node ( keys %$cdn ) {
		# [en]	check if cdn node is in out-of-service mode
		# [de]	Erstmal pruefen ob dieser Server ausser Betrieb ist
		if ( $cdn->{ $node }->{'disabled'} ) {
			$debug && print STDERR "DEBUG\tcdn node $node is out-of-service\n";
			next;
		}

		# [en]	get geo-location for CDN node from stored values (preferred)
		#	or run a geo-lookup (slower)
		# [de]	ermittle die Geo-Informationen fuer den CDN-Server aus gespeicherten
		# 	Werten (bevorzugt/schneller) oder ueber Geo-Aufloesung (langsamer)
		if ( ! (    $cdn->{ $node }->{'country'}
			and $cdn->{ $node }->{'latitude'}
			and $cdn->{ $node }->{'longitude'} ) ) {

			( $cdn->{ $node }->{'country'},
			  $cdn->{ $node }->{'latitude'},
			  $cdn->{ $node }->{'longitude'} ) =
				get_country_for( $cdn->{ $node }->{'ipv4'} );
		}

		$debug && printf STDERR "DEBUG\tdistance between %0.2f/%0.2f and %0.2f/%0.2f\n",
			$latitude, $longitude, $cdn->{ $node }->{'latitude'},
			$cdn->{ $node }->{'longitude'};
		my $distance = distance( $latitude, $longitude,
			$cdn->{ $node }->{'latitude'},
			$cdn->{ $node }->{'longitude'}
		);

		$debug && printf STDERR "DEBUG\tdistance between %s and %s is %5d km\n",
			uc $country, uc $cdn->{ $node }->{'country'},
			$distance;

		# [en]	store smallest distance
		# [de]	merke die kuerzeste Entfernung
		if ( ! $pref->{'location'} 	# [en]	no preferred location
					 	# [de]	keine bevorzugte Position
		     or $distance < $pref->{'distance'} ) {
			$pref->{'location'} = $cdn->{ $node }->{'country'};
			$pref->{'hostname'} = $node;
			$pref->{'distance'} = $distance;
		}

	}
	$debug && printf STDERR ( "DEBUG\ttraffic is handeled by CDN node %s (in %s)\n",
		$pref->{'hostname'}, uc $pref->{'location'} );

	return $pref->{'hostname'};
}



# [en]	the following subroutines are based on sample code by
#	http://www.geodatasource.com/developers/perl
#	which is licensed under LGPLv3
# [de]	Die folgenden Funktionen basieren auf Beispielcode von
#	http://www.geodatasource.com/developers/perl
#	und sind lizensiert unter der LGPLv3

#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
#::: [en] This routine calculates the distance between two points (given the  :::
#:::      latitude/longitude of those points).                                :::
#::: [de] Diese Funktion berechnet die Entfernung zwischen zwei Punkten (von  :::
#:::      den Punkten sind die Laengen- und Breitengrade bekannt).            :::
#::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
sub distance ($$$) {
	my ($lat1, $lon1, $lat2, $lon2) = @_;
	my $theta = ( $lon1 - $lon2 ) * PI_180;
	$lat1 *= PI_180;	# [en] converts decimal degrees to radians
	$lat2 *= PI_180;	# [de] Konvertiere Radian zu dezimalen Graden

	my $dist = sin( $lat1 ) * sin( $lat2 )
		 + cos( $lat1 ) * cos( $lat2 )
		 * cos( $theta );

	$dist  = atan2( sqrt( 1 - $dist**2 ), $dist );
	return ($dist);
}
